package ap.parser;

import ap.terfor.{ConstantTerm, OneTerm, TermOrder}
import ap.terfor.conjunctions.{Conjunction, Quantifier}
import ap.terfor.linearcombination.LinearCombination
import ap.terfor.equations.{EquationConj, NegEquationConj}
import ap.terfor.inequalities.InEqConj
import ap.terfor.preds.{Predicate, Atom}
import ap.util.{Debug, Logic, PlainRange}
import ap.basetypes.IdealInt
import ApInput._
import ApInput.Absyn._

import scala.collection.mutable.{Stack, ArrayBuffer}

object Parser2InputAbsy {

  private val AC = Debug.AC_PARSER
  
  class ParseException(msg : String) extends Exception(msg)
  class TranslationException(msg : String) extends Exception(msg)
  
  /**
   * Parse a problem from a character stream. The result is the formula
   * contained in the input, the set of constants that are to be eliminated,
   * and the <code>TermOrder</code> that was used for the formula. It is
   * guaranteed that the eliminated constants are maximal in the term order.
   */
  def apply(input : java.io.Reader) : (IFormula, Set[ConstantTerm], TermOrder) = {
    val l = new Yylex(new CRRemover2 (input))
    val p = new parser(l)
    
    val api = try { p.pAPI } catch {
      case e : Exception =>
        throw new ParseException("At line " + String.valueOf(l.line_num()) +
                                 ", near \"" + l.buff() + "\" :" +
                                 "     " + e.getMessage())
    }
    
    apply(api)
  }
  
  def apply(api : API) : (IFormula, Set[ConstantTerm], TermOrder) = {
    val t = new Parser2InputAbsy
    t collectDeclarations api
    t translateProblem api
  }
  
}


/**
 * Class for removing all CR-characters in a stream (necessary because the
 * lexer seems to dislike CRs in comments)
 */
private class CRRemover2(input : java.io.Reader) extends java.io.Reader {
  
  private val CR : Int = '\r'
   
  def read(cbuf : Array[Char], off : Int, len : Int) : Int = {
    var read = 0
    while (read < len) {
      val next = input.read
      next match {
        case -1 => return if (read == 0) -1 else read
        case CR => // nothing, read next character
        case _ => {
          cbuf(off + read) = next.toChar
          read = read + 1
        }
      }
    }
    read
  }
   
  def close : Unit = input.close
  
}


class Parser2InputAbsy private {
  
  private val env = new Environment
  
  import IExpression._
  
  /** Implicit conversion so that we can get a scala-like iterator from a
    * a Java list */
  private implicit def toWrappedList[A](list : java.util.LinkedList[A]) =
    new scala.collection.jcl.LinkedList(list)
  
  //////////////////////////////////////////////////////////////////////////////
    
  private def collectDeclarations(api : API) : Unit = api match {
    case api : BlockList =>
      for (block <- api.listblock_.elements) block match {
        case block : FunctionDecls =>
          for (decl <- block.listdeclfunc_.elements)
            collectDeclFunC(decl,
                            (id) => env.addConstant(new ConstantTerm(id),
                                                    Environment.Universal))
        case block : ExConstants =>
          for (decl <- block.listdeclexconstantc_.elements)
            collectDeclExConstantC(decl,
                                   (id) => env.addConstant(new ConstantTerm(id),
                                                           Environment.Existential))
        case block : PredDecls =>
          for (decl <- block.listdeclpredc_.elements) decl match {
            case decl : DeclPred => {
              val name = decl.ident_
              val arity = decl.optformalargs_ match {
                case _ : NoFormalArgs => 0
                case args : WithFormalArgs => determineArity(args.formalargsc_)
              }
              env.addPredicate(new Predicate(name, arity))
            }
          }
        case _ : Problem => /* nothing */
      }
  }

  private def determineArity(args : FormalArgsC) : Int = args match {
    case args : FormalArgs => {
      //////////////////////////////////////////////////////////////////////////
      Debug.assertInt(Parser2InputAbsy.AC,
                      Logic.forall(for (at <- args.listargtypec_.elements)
                                   yield (at.asInstanceOf[ArgType].type_.isInstanceOf[TypeInt])))
      //////////////////////////////////////////////////////////////////////////
      args.listargtypec_.size
    }
  }
  
  //////////////////////////////////////////////////////////////////////////////

  private def collectDeclFunC(decl : DeclFunC, addCmd : String => unit) : Unit =
    decl match {
      case decl : DeclFun => {
        ////////////////////////////////////////////////////////////////////////
        Debug.assertInt(Parser2InputAbsy.AC, decl.type_.isInstanceOf[TypeInt])
        ////////////////////////////////////////////////////////////////////////
        val partial = toWrappedList(decl.listfunoption_) match {
          case Seq() => false
          case Seq(_ : Partial) => true
          case _ => {
            val strs = for (o <- decl.listfunoption_) yield funOption2String(o)
            throw new Parser2InputAbsy.TranslationException(
                       "Illegal options for function: " + (strs mkString " "))
          }
        }
        env.addFunction(new IFunction (decl.ident_,
                                       determineArity(decl.formalargsc_),
                                       partial))
      }
      case decl : DeclUniConstant => {
        if (!decl.listfunoption_.isEmpty)
          throw new Parser2InputAbsy.TranslationException(
                                        "Constants do not have options")
        collectDeclarations(decl.declvarconstc_, addCmd)
      }
    }
    
  private def funOption2String(option : FunOption) : String = option match {
    case _ : Partial => "\\partial"
  }
  
  private def collectDeclExConstantC(decl : DeclExConstantC, addCmd : String => Unit) : Unit =
    collectDeclarations(decl.asInstanceOf[DeclExConstant].declvarconstc_, addCmd)

  private def collectDeclBinder(decl : DeclBinder,
                                addCmd : String => unit) : Unit = decl match {
    case decl : DeclBinder1 => collectDeclarations(decl.declvarconstc_, addCmd)
    case decl : DeclBinderM => for (decl <- decl.listdeclvarconstc_.elements) 
                                 collectDeclarations(decl, addCmd)
  }

  private def collectDeclarations(decl : DeclVarConstC,
                                  addCmd : String => Unit) : Unit = decl match {
    case decl : DeclVarConst => { 
      //////////////////////////////////////////////////////////////////////////
      Debug.assertInt(Parser2InputAbsy.AC, decl.type_.isInstanceOf[TypeInt])
      //////////////////////////////////////////////////////////////////////////
      for (id <- decl.listident_.elements) addCmd(id)
    }
  }

  //////////////////////////////////////////////////////////////////////////////
  
  def translateProblem(api : API)
              : (IFormula, Set[ConstantTerm], TermOrder) = api match {
    case api : BlockList => {
      api.listblock_.filter(_.isInstanceOf[Problem]) match {
        case Seq(problem : Problem) => {
          val f = asFormula(translateExpression(problem.expression_))
          (f, env.universalConstants, env.order)
        }
        case _ => throw new Parser2InputAbsy.TranslationException(
                             "Found zero or more than one \\problem blocks")
      }
    }
  }

  //////////////////////////////////////////////////////////////////////////////

  private def translateExpression(f : Expression) : IExpression = f match {
    ////////////////////////////////////////////////////////////////////////////
    // Formulae
    case f : ExprEqv =>
      translateBinForConnective(f.expression_1, f.expression_2, _ <-> _)
    case f : ExprImp =>
      translateBinForConnective(f.expression_1, f.expression_2, _ -> _)
    case f : ExprOr => {
      val subs = collectSubExpressions(f, _.isInstanceOf[ExprOr])
      (for (f <- subs.elements)
         yield asFormula(translateExpression(f))) reduceLeft (_ | _)
    }
    case f : ExprAnd => {
      val subs = collectSubExpressions(f, _.isInstanceOf[ExprAnd])
      (for (f <- subs.elements)
         yield asFormula(translateExpression(f))) reduceLeft (_ & _)
    }
    case f : ExprNot =>
      translateUnForConnective(f.expression_, ! _)
    case f : ExprQuant =>
      translateQuant(f)
    case _ : ExprTrue => i(true)
    case _ : ExprFalse => i(false)
    case f : ExprIdApp => translateExprIdApp(f)
    case f : ExprRel =>
      translateBinTerConnective(f.expression_1, f.expression_2,
                                f.relsym_ match {
                                  case _ : RelEq => _ === _
                                  case _ : RelNEq => _ =/= _
                                  case _ : RelLeq => _ <= _
                                  case _ : RelGeq => _ >= _
                                  case _ : RelLt => _ < _
                                  case _ : RelGt => _ > _
                                })
    case f : ExprTrigger => translateTrigger(f)
    ////////////////////////////////////////////////////////////////////////////
    // Terms
    case t : ExprPlus =>
      translateBinTerConnective(t.expression_1, t.expression_2, _ + _)
    case t : ExprMinus =>
      translateBinTerConnective(t.expression_1, t.expression_2, _ - _)
    case t : ExprMult =>
      translateBinTerConnective(t.expression_1, t.expression_2, mult _)
    case t : ExprUnPlus =>
      translateUnTerConnective(t.expression_, (lc) => lc)
    case t : ExprUnMinus =>
      translateUnTerConnective(t.expression_, - _)
    case t : ExprLit =>
      IIntLit(IdealInt(t.intlit_))
  }
  
  private def collectSubExpressions(f : Expression, cont : Expression => Boolean)
                                                            : Seq[Expression] = {
    val todo = new Stack[Expression]
    val res = new ArrayBuffer[Expression]
    
    todo push f
    while (!todo.isEmpty) {
      val next = todo.pop
      
      if (cont(next)) {
        next match {
          case BinConnective(f1, f2) => { todo push f1; todo push f2 }
        }
      } else {
        res += next
      }
    }
    
    res
  }
  
  private object BinConnective {
    def unapply(f : Expression) : Option[(Expression, Expression)] = f match {
      case f : ExprAnd => Some((f.expression_1, f.expression_2))
      case f : ExprOr => Some((f.expression_1, f.expression_2))
    }
  }
  
  private def asFormula(expr : IExpression) : IFormula = expr match {
    case expr : IFormula =>
      expr
    case _ => 
      throw new Parser2InputAbsy.TranslationException("Expected a formula, not " + expr)
  }
  
  private def asTerm(expr : IExpression) : ITerm = expr match {
    case expr : ITerm =>
      expr
    case _ => 
      throw new Parser2InputAbsy.TranslationException("Expected a term, not " + expr)
  }
  
  private def translateUnForConnective(f : Expression, con : (IFormula) => IExpression)
              : IExpression =
    con(asFormula(translateExpression(f)))
  
  private def translateUnTerConnective(f : Expression, con : (ITerm) => IExpression)
              : IExpression =
    con(asTerm(translateExpression(f)))
  
  private def translateBinForConnective(f1 : Expression, f2 : Expression,
                                        con : (IFormula, IFormula) => IExpression)
              : IExpression =
    con(asFormula(translateExpression(f1)), asFormula(translateExpression(f2)))
  
  private def translateBinTerConnective(f1 : Expression, f2 : Expression,
                                        con : (ITerm, ITerm) => IExpression)
              : IExpression =
    con(asTerm(translateExpression(f1)), asTerm(translateExpression(f2)))
  
  private def translateQuant(f : ExprQuant) : IFormula = {
    val quant : Quantifier = f.quant_ match {
      case _ : QuantAll => Quantifier.ALL
      case _ : QuantEx => Quantifier.EX
    }
    
    // add the bound variables to the environment and record their number
    var quantNum : Int = 0
    collectDeclBinder(f.declbinder_,
                      (id) => { quantNum = quantNum + 1; env pushVar id })
    
    val res = translateUnForConnective(f.expression_,
                                       quan(Array.make(quantNum, quant), _))

    // pop the variables from the environment
    for (_ <- PlainRange(quantNum)) env.popVar
    
    res.asInstanceOf[IFormula]
  }

  private def translateExprIdApp(f : ExprIdApp) : IExpression =
    env.lookupSym(f.ident_) match {
      case Environment.Predicate(pred) => {
        val args = translateOptArgs(f.optargs_)
        if (pred.arity != args.size)
          throw new Parser2InputAbsy.TranslationException(
              "Predicate " + pred +
              " is applied to a wrong number of arguments: " + (args mkString ", "))
        
        IAtom(pred, args)
      }
      
      case Environment.Function(fun) => {
        val args = translateOptArgs(f.optargs_)
        if (fun.arity != args.size)
          throw new Parser2InputAbsy.TranslationException(
              "Function " + fun +
              " is applied to a wrong number of arguments: " + (args mkString ", "))
        
        IFunApp(fun, args)
      }
      
      case Environment.Constant(c, _) => {
        f.optargs_ match {
          case _ : Args =>
            throw new Parser2InputAbsy.TranslationException(
                               "Constant " + c + " does not have arguments")
          case _ : NoArgs => // nothing
        }
        c
      }
      
      case Environment.Variable(i) => {
        f.optargs_ match {
          case _ : Args =>
            throw new Parser2InputAbsy.TranslationException(
                               "Variable " + f.ident_ + " does not have arguments")
          case _ : NoArgs => // nothing
        }
        v(i)
      }
    }
  
  private def translateTrigger(trigger : ExprTrigger) :IFormula = {
    val patterns = translateArgs(trigger.listargc_)
    val body = asFormula(translateExpression(trigger.expression_))
    ITrigger(patterns, body)
  }
  
  private def translateOptArgs(args : OptArgs) = args match {
    case args : Args => translateArgs(args.listargc_)
    case _ : NoArgs => List()
  }
  
  private def translateArgs(args : ListArgC) =
    for (arg <- args) yield arg match {
      case arg : Arg => asTerm(translateExpression(arg.expression_))
    }

  //////////////////////////////////////////////////////////////////////////////

  private def mult(t1 : ITerm, t2 : ITerm) : ITerm =
    try { t1 * t2 }
    catch {
      case _ : IllegalArgumentException =>
        throw new Parser2InputAbsy.TranslationException(
                        "Do not know how to multiply " + t1 + " with " + t2)
    }

}
