package ap.proof

import ap.terfor.ConstantTerm
import ap.terfor.conjunctions.{Conjunction, Quantifier}
import ap.terfor.linearcombination.LinearCombination
import ap.util.{Debug, Logic, FilterIt}

object ConstantFreedom {
  
  private val AC = Debug.AC_UNIFICATION_OPT
  
  abstract class ConstantStatus extends PartiallyOrdered[ConstantStatus] {
    def meet(that : ConstantStatus) : ConstantStatus =
      if (this <= that) this else that

    def tryCompareTo[B >: ConstantStatus <% PartiallyOrdered[B]](that : B)
                                                               : Option[Int] =
      (this, that) match {
        case (NonFree, ShieldingEquations) => Some(-1)
        case (ShieldingEquations, NonFree) => Some(1)
        case _ => Some(0) // then this == that
      }
  }
  
  case object NonFree extends ConstantStatus
  case object ShieldingEquations extends ConstantStatus
  
  val BOTTOM = new ConstantFreedom (Map() withDefaultValue NonFree)
  
}

/**
 * Class to represent the set of constants that are considered as "free" (i.e.,
 * that are only unifiable with themselves). Objects of this class are stored
 * in the nodes of proof tree and are updated when the proof is expanded in
 * order to eventually reach a fixed point.
 * 
 * TODO: avoid the creation of new objects whenever possible
 */
class ConstantFreedom private (private val constantStatus :
                                 Map[ConstantTerm, ConstantFreedom.ConstantStatus])
      extends PartiallyOrdered[ConstantFreedom] {

  import ConstantFreedom.{NonFree, ShieldingEquations, ConstantStatus}
  
  //////////////////////////////////////////////////////////////////////////////
  // The status map only explicitly mentions the free constants. All other
  // constants are supposed to be mapped to NonFree by the default-method
  // of the map
  Debug.assertCtor(ConstantFreedom.AC,
                   !(constantStatus.values contains NonFree) &&
                   constantStatus(new ConstantTerm ("X")) == NonFree)
  //////////////////////////////////////////////////////////////////////////////

  override def <=[B >: ConstantFreedom <% PartiallyOrdered[B]](that : B) : Boolean =
    that match {
      case that : ConstantFreedom =>
        this.constantStatus.size <= that.constantStatus.size &&
        Logic.forall(for ((c, s) <- this.constantStatus.elements)
                     yield (s <= that.constantStatus(c)))
      case _ => false
    }
    
  def tryCompareTo[B >: ConstantFreedom <% PartiallyOrdered[B]](that : B)
                                                              : Option[Int] =
    that match {
      case that : ConstantFreedom => if (this == that)
                                       Some(0)
                                     else if (this <= that)
                                       Some(-1)
                                     else if (that <= this)
                                       Some(1)
                                     else
                                       None
      case _ => None
    }

  def isBottom : Boolean = constantStatus.isEmpty
                                                              
  def meet(that : ConstantFreedom) : ConstantFreedom = {
    var newStatus = this.constantStatus
    for ((c, s) <- this.constantStatus) {
      (s meet that.constantStatus(c)) match {
        case NonFree => newStatus = newStatus - c
        case newS if (newS != s) => newStatus = newStatus + (c -> newS)
        case _ => // nothing
      }
    }
    new ConstantFreedom(newStatus)
  }

  /**
   * Set the status of the given constants to the top value (as free as possible)
   */
  def addTopStatus(consts : Iterable[ConstantTerm]) : ConstantFreedom =
    new ConstantFreedom(this.constantStatus ++
                        (for (c <- consts.elements) yield (c -> ShieldingEquations)))
  
  def --(consts : Iterable[ConstantTerm]) : ConstantFreedom =
    new ConstantFreedom(this.constantStatus -- consts)
  
  /**
   * Given a constraint from which all shielded parts have been removed,
   * determine which constants have to be considered as non-free
   */
  def findNonFreeness(unshieldedConj : Conjunction, bc : BindingContext) : ConstantFreedom =
    new ConstantFreedom(this.constantStatus -- unshieldedConj.constants)
  
  /**
   * List all constants whose status is different in <code>this</code> and
   * <code>that</code>
   */
  def changedConstants(that : ConstantFreedom) : Set[ConstantTerm] =
    (Set() ++ this.constantStatus.keySet ++ that.constantStatus.keySet)
             .filter ((c) => this.constantStatus(c) != that.constantStatus(c))
    
  //////////////////////////////////////////////////////////////////////////////
                                                              
  override def equals(that : Any) : Boolean =
    that match {
      case that : ConstantFreedom => this.constantStatus == that.constantStatus
      case _ => false
    }
  
  override def hashCode : Int = this.constantStatus.hashCode
  
  override def toString : String = this.constantStatus.toString
  
  //////////////////////////////////////////////////////////////////////////////
  
  /**
   * Determine whether <code>c</code> is a formula of the shape
   * <code> a*x + t = 0 & phi </code>, where <code>a != 0</code> and
   * <code>x</code> is a quasi-universal constant that is maximal in
   * <code> a*x + t = 0 </code>.
   */
  def isShielded(c : Conjunction, bc : BindingContext) : Boolean =
    Logic.exists(for (lc <- c.arithConj.positiveEqs.elements)
                 yield isShieldingLC(lc, bc)) ||
    Logic.exists(for (lc <- c.arithConj.inEqs.equalityInfs.elements)
                 yield isShieldingLC(lc, bc)) ||
    Logic.exists(for (negConj <- c.negatedConjs.elements)
                 yield Logic.forall(for (d <- negConj.elements)
                                    yield isShieldedNeg(d, bc)))
  
  private def isShieldingLC(lc : LinearCombination, bc : BindingContext) : Boolean =
    lc.variables.isEmpty &&
    Logic.exists(for (c <- (bc maximumConstants lc.constants).elements)
                 yield (constantStatus(c) != NonFree))

  private def isShieldedNeg(c : Conjunction, bc : BindingContext) : Boolean =
    isShielded(c.negate, bc)

  /**
   * Determine the (disjunctively connected) unshielded part of a formula
   */
  def unshieldedPart(c : Conjunction, bc : BindingContext) : Conjunction =
    if (!c.quans.isEmpty)
      c
    else if (isShielded(c, bc))
      Conjunction.FALSE
    else if (c.isNegatedConjunction)
      Conjunction.conj(FilterIt(c.negatedConjs(0).elements,
                                (d:Conjunction) => !isShieldedNeg(d, bc)),
                       c.order).negate
    else
      // TODO: generalise?
      c

  //////////////////////////////////////////////////////////////////////////////
  
  /**
   * Only used for runtime assertion purposes
   */
  protected[proof] def freeConstsAreUniversal(bc : BindingContext) : Boolean =
    Logic.forall(for (c <- constantStatus.keys)
                 yield ((bc binder c) == Some(Quantifier.ALL)))
}