package ap.proof;

import ap.terfor.{Formula, TermOrder}
import ap.terfor.conjunctions.{Conjunction, Quantifier}
import ap.proof.goal._
import ap.util.{Logic, Debug}
import ap.parameters.{GoalSettings, Param}

object ExhaustiveProver {
  
  private def AC = Debug.AC_PROVER
  
}

/**
 * A prover that tries to construct an exhaustive proof for a given goal. The
 * prover tries to optimise by early stopping the expansion of the proof tree
 * if it is detected that a certain subtree can never yield a satisfiable
 * closing constraint. There are two main modes of operation: with
 * <code>depthFirst</code>, it is tried to derive a satisfiable constraint from
 * the given problem, without aiming for exhaustiveness. Without this option,
 * the tree is expanded depth-first until it is exhaustive (which terminates
 * in the case of PA formulae, but not in general).
 */
class ExhaustiveProver(simplifier : ConstraintSimplifier,
                       depthFirst : Boolean,
                       settings : GoalSettings) {

  def this(simplifier : ConstraintSimplifier) =
    this(simplifier, false, GoalSettings.DEFAULT)

  private def ptfStoppingCond(goal : Goal) = {
    Timeout.check
    (goal.tasks.max match {
 //   case FactsNormalisationTask | EliminateFactsTask |
 //        _ : AddFactsTask | _ : NegLitClauseTask => false
      case FactsNormalisationTask | EliminateFactsTask | UpdateTasksTask => false
      case task : AddFactsTask => Param.POS_UNIT_RESOLUTION(goal.settings) &&
                                  !task.formula.predicates.isEmpty
      case task : BetaFormulaTask => task.splittingNecessary(goal)
      case task : DivisibilityTask => task.splittingNecessary(goal)
      case SplitEqTask => SplitEqTask.splittingNecessary(goal)
      case OmegaTask => OmegaTask.splittingNecessary(goal)
      case _ => true
    })
  }

  private val ptf =
    new IteratingProofTreeFactory(ptfStoppingCond _, simplifier)
   
  //////////////////////////////////////////////////////////////////////////////

  def apply(inputFor : Formula, order : TermOrder) : ProofTree = {
    ////////////////////////////////////////////////////////////////////////////
    Debug.assertPre(ExhaustiveProver.AC,
                    inputFor.variables.isEmpty && (order isSortingOf inputFor))
    ////////////////////////////////////////////////////////////////////////////
    val goal = Goal(inputFor, order, settings)
    Timeout.unfinished {
      
      (if (depthFirst)
         expandDepthFirstUntilSat(goal, false)
       else
         expandFairUntilSat(goal, false)) _1
        
    } {
      case None => goal
      case x => x
    }
  }

  private def continueProving(tree : ProofTree) = {
    Timeout.check
    (tree.stepMeaningful || !tree.fixedConstantFreedom) &&
    // when proving in depth-first mode, a tree is expanded only as long as
    // its constraint is not satisfiable
    (!depthFirst || ModelSearchProver(tree.closingConstraint.negate,
                                      tree.closingConstraint.order).isFalse)
  }
  
  //////////////////////////////////////////////////////////////////////////////

  /**
   * Expand the proof breadth-first until the root constraint is satisfiable
   * (or until <code>stoppingCond</code> returns <code>true</code>).
   * The result is a pair consisting of the new proof tree and a boolean
   * that tells whether it was possible to apply any steps
   */
  private def expandFairUntilSat(_tree : ProofTree, swpBefore : Boolean)
                                                     : (ProofTree, Boolean) = {
    var tree : ProofTree = _tree
    var cont : Boolean = true
    var swp : Boolean = swpBefore
    
    Timeout.unfinished {
      // if a timeout occurs, we return the proof tree that has been constructed
      // up to this point
      
      while (cont && continueProving(tree)) {
  /*  println(tree)
     println(goalNum(tree))
     println  */
        val (newTree, newCont) =
          expandProofGoals(tree, tree.closingConstantFreedom, 1)
        tree = newTree
        cont = newCont
        if (newCont) swp = true
      }
      
    } {
      case _ => tree
    }
    
    (tree, swp)
  }

  private def isGoalLike(tree : ProofTree) : Boolean = tree match {
    case _ : Goal                                   => true
    case QuantifiedTree(Quantifier.ALL, _, subtree) => isGoalLike(subtree)
    case _                                          => false
  }

  /**
   * Expand the proof depth-first until the root constraint is satisfiable
   * (or until <code>stoppingCond</code> returns <code>true</code>).
   * The result is a pair consisting of the new proof tree and a boolean
   * that tells whether it was possible to apply any steps
   */
  private def expandDepthFirstUntilSat(tree : ProofTree, swpBefore : Boolean)
                                                       : (ProofTree, Boolean) = {
    ////////////////////////////////////////////////////////////////////////////
    // if there are free constants, we have to expand in a fair manner
    Debug.assertPre(ExhaustiveProver.AC, tree.constantFreedom.isBottom)
    ////////////////////////////////////////////////////////////////////////////
    if (Timeout.unfinishedValue(tree) { continueProving(tree) })
      tree match {
      case _ if (isGoalLike(tree) ||
                 tree.subtrees.exists((subtree) =>
                                      !subtree.constantFreedom.isBottom)) => {
        // we always consider the goal together with the enclosing universal
        // quantifiers
        val (newTree, swp) =
          Timeout.unfinishedValue(tree) {
            expandProofGoals(tree, ConstantFreedom.BOTTOM, 1)
          }
        
        if (swp) {
          // recurse because the structure of the tree might have changed
          expandDepthFirstUntilSat(newTree, true)
        } else {
          // we continue in a fair manner
          expandFairUntilSat(newTree, swpBefore)
        }
      }
      
      case goal : Goal => {
        val goalOneStep = Timeout.unfinishedValue(goal) { goal step ptf }
        expandDepthFirstUntilSat(goalOneStep, true)
      }
      
      case tree : WeakenTree => // should not happen, in this case the constraint
                                // of the proof tree would be satisfiable
                                { assert(false); null }
      
      case tree : QuantifiedTree => {
        val (newSubtree, swp) =
          Timeout.unfinished {
            expandDepthFirstUntilSat(tree.subtree, false)
          } {
            case lastTree : ProofTree =>
              tree.update(lastTree, ConstantFreedom.BOTTOM)
          }
        val newTree = tree.update(newSubtree, ConstantFreedom.BOTTOM)
        
        if (swp) {
          // recurse because the structure of the tree might have changed
          expandDepthFirstUntilSat(newTree, true)
        } else {
          // we continue in a fair manner
          expandFairUntilSat(newTree, swpBefore)
        }        
      }
      
      case tree : AndTree => {
        val (newLeft, leftSWP) =
          Timeout.unfinished {
            expandDepthFirstUntilSat(tree.left, false)
          } {
            case lastTree : ProofTree =>
              tree.update(lastTree, tree.right, ConstantFreedom.BOTTOM)
          }
        val (newRight, rightSWP) =
          Timeout.unfinished {
            expandDepthFirstUntilSat(tree.right, false)
          } {
            case lastTree : ProofTree =>
              tree.update(newLeft, lastTree, ConstantFreedom.BOTTOM)
          }
        
        val newTree = tree.update(newLeft, newRight, ConstantFreedom.BOTTOM)
        
        if (leftSWP || rightSWP) {
          // recurse because the structure of the tree might have changed
          expandDepthFirstUntilSat(newTree, true)
        } else {
          // we continue in a fair manner
          expandFairUntilSat(newTree, swpBefore)
        }
      }
    } else {
      // otherwise, the tree is already satisfiable (or at least the constraint
      // is not false .. TODO) or no more steps are possible
      (tree, swpBefore)
    }
  }
  
  /**
   * Try to expand each goal of <code>tree</code> by applying <code>limit</code>
   * steps. The result is a pair consisting of the new proof tree and a boolean
   * that tells whether it was possible to apply any steps
   */
  private def expandProofGoals(tree : ProofTree,
                               newConstantFreedom : ConstantFreedom,
                               limit : Int) : (ProofTree, Boolean) =
    if (limit > 0 &&
        (tree.stepMeaningful && !tree.closingConstraint.isTrue ||
         newConstantFreedom != tree.constantFreedom || !tree.fixedConstantFreedom))
      tree match {
      
      case goal : Goal => {
        val updatedGoal = goal updateConstantFreedom newConstantFreedom
        val newTree =
          if (updatedGoal.stepPossible)
            expandProofGoals(updatedGoal.step(ptf), newConstantFreedom, limit - 1) _1
          else
            updatedGoal
        (newTree, true)  // also updating the set of free constants is
                         // considered as a proof step
      }
      
      case tree : ProofTreeOneChild => {
        val (newSubtree, stepWasPossible) =
          expandProofGoals(tree.subtree,
                           tree newConstantFreedomForSubtree newConstantFreedom,
                           limit)
        (tree.update(newSubtree, newConstantFreedom), stepWasPossible)
      }
      
      case tree : AndTree => {
        val (leftCF, rightCF) = tree newConstantFreedomForSubtrees newConstantFreedom
        val (newLeft, leftSWP) = expandProofGoals(tree.left, leftCF, limit)
        val (newRight, rightSWP) = expandProofGoals(tree.right, rightCF, limit)        
        (tree.update(newLeft, newRight, newConstantFreedom), leftSWP || rightSWP)
      }
      
    } else {
      (tree, false)
    }
   
   private def goalNum(tree : ProofTree) : Int = tree match {
     case g : Goal => if (g.closingConstraint.isTrue) 0 else 1
     case t : ProofTreeOneChild => goalNum(t.subtree)
     case t : AndTree => goalNum(t.left) + goalNum(t.right)
   }
}
