package ap.proof

import ap.terfor.{Formula, TermOrder}
import ap.terfor.conjunctions.{Conjunction, Quantifier}
import ap.proof.goal._
import ap.util.{Debug, Seqs}
import ap.parameters.{GoalSettings, Param}

/**
 * Prover to eliminate quantifiers in a PA formula. The prover is losely
 * based on the idea in the paper
 *   "A Quantifier Elimination Algorithm for Linear Real Arithmetic"
 * by David Monniaux, although it does not explicitly compute solutions
 * of the matrix of a quantified formula like in the paper. Instead, the
 * constraint extracted from an exhaustive subtree is injected as a lemma
 * into other subtrees and used to close such subtrees earlier.
 */
object QuantifierElimProver {

  private val AC = Debug.AC_PROVER

  private val simplifier = ConstraintSimplifier.NO_SIMPLIFIER
  
  private val ptf = new SimpleProofTreeFactory(true, simplifier)
  private val nonRemovingPtf = new SimpleProofTreeFactory(false, simplifier)

  private val completeCNFSettings =
    Param.FULL_SPLITTING.set(GoalSettings.DEFAULT, true)
  private val nonCompleteCNFSettings =
    Param.FULL_SPLITTING.set(GoalSettings.DEFAULT, false)
  
  //////////////////////////////////////////////////////////////////////////////
  
  def apply(inputFor : Formula,
            // fully transform the given formula into CNF,
            // instead of leaving subformulae that are not relevant
            // for quantifier elimination unexpanded
            completeCNF : Boolean,
            order : TermOrder) : Conjunction = {
    ////////////////////////////////////////////////////////////////////////////
    Debug.assertPre(QuantifierElimProver.AC,
                    ((Conjunction collectQuantifiers inputFor) subsetOf Set(Quantifier.ALL)) &&
                    (order isSortingOf inputFor))
    ////////////////////////////////////////////////////////////////////////////
    val settings = if (completeCNF) completeCNFSettings else nonCompleteCNFSettings
    val goal = Goal(inputFor, order, settings)
    expandProof(goal, Conjunction.FALSE, 0)
  }

  private def checkPruning(goal : Goal) =
    goal.tasks.max match {
      case task : BetaFormulaTask => task.splittingNecessary(goal)
      case SplitEqTask => SplitEqTask.splittingNecessary(goal)
      case EliminateFactsTask => true
      case _ => false
    }

  private def expandProof(tree : ProofTree,
                          // if the following formula can be reduced to true,
                          // proving on this branch can be stopped
                          pruningFor : Conjunction,
                          depth : Int) : Conjunction = {
    Timeout.check
    
    if (!tree.stepPossible) {
      tree.closingConstraint
 //     println("backtracking (depth " + depth + ")")
    } else tree match { 
      case goal : Goal =>
        if (checkPruning(goal)) {
          //////////////////////////////////////////////////////////////////////
          Debug.assertInt(QuantifierElimProver.AC,
                          // check that no defined symbols occur in the
                          // pruning criterion
                          (goal definedSyms pruningFor) == pruningFor)
          //////////////////////////////////////////////////////////////////////
          val newPruningFor = goal definedSyms (goal reduceWithFacts pruningFor)
          if (newPruningFor.isTrue)
            Conjunction.TRUE
          else
            expandProof(goal.step(ptf), newPruningFor, depth)
        } else {
          expandProof(goal.step(ptf), pruningFor, depth)            
        }

      case WeakenTree(disjunct, subtree) =>
        Conjunction.disj(Array(expandProof(subtree, pruningFor, depth), disjunct),
                         tree.order)

      case QuantifiedTree(Quantifier.ALL, consts, subtree) => {
        // quantifiers can be ignored, because it is assumed that eliminated
        // constants actually are eliminated and do not occur in constraints
        val res = expandProof(subtree, pruningFor, depth)
        ////////////////////////////////////////////////////////////////////////
        Debug.assertPre(QuantifierElimProver.AC,
                        Seqs.disjoint(res.constants, consts))
        ////////////////////////////////////////////////////////////////////////
        res
      }

      case tree : AndTree => {
        val leftRes =
          expandProof(tree.left, pruningFor, depth + 1)
        val resAndPruningFor =
          Conjunction.disj(Array(pruningFor, leftRes.negate), tree.order)
        val rightRes = expandProof(tree.right, resAndPruningFor, depth + 1)
        Conjunction.conj(Array(leftRes, rightRes), tree.order)
      }
      
    }
  }
  
}
