package ap.proof;

import ap.terfor.{Formula, TermOrder}
import ap.terfor.conjunctions.Conjunction
import ap.util.Debug

object WeakenTree {
  
  private val AC = Debug.AC_PROOF_TREE

  def apply(subtree : ProofTree,
            disjunct : Conjunction,
            vocabulary : Vocabulary) : ProofTree =
    if (disjunct.isFalse) {
      subtree
    } else subtree match {
      case subtree : WeakenTree =>
        WeakenTree(subtree.subtree,
                   Conjunction.disj(Array(subtree.disjunct, disjunct),
                                    vocabulary.order),
                   vocabulary)
      case _ => new WeakenTree(subtree, disjunct, vocabulary)
    }
  
  def unapply(p : ProofTree) : Option[(Conjunction, ProofTree)] = p match {
    case p : WeakenTree => Some((p.disjunct, p.subtree))
    case _ => None
  }
  
}

/**
 * <code>ProofTreeOneChild</code> that weakens the closing constraint of its
 * <code>subtree</code> by disjunctively adding a formula
 */
class WeakenTree private (val subtree : ProofTree,
                          val disjunct : Conjunction,
                          val vocabulary : Vocabulary)
      extends { protected val subtreeOrder : TermOrder = vocabulary.order }
              with ProofTreeOneChild {
  
  //////////////////////////////////////////////////////////////////////////////
  Debug.assertCtor(WeakenTree.AC,
                   (order isSortingOf disjunct) &&
                   !subtree.isInstanceOf[WeakenTree])
  //////////////////////////////////////////////////////////////////////////////

  private lazy val unshieldedDisjunct =
    constantFreedom.unshieldedPart(disjunct, bindingContext) 
  
  lazy val closingConstraint : Conjunction =
    Conjunction.disj(Array(subtree.closingConstraint, unshieldedDisjunct),
                     subtreeOrder)
    
  lazy val closingConstantFreedom : ConstantFreedom =
    subtree.closingConstantFreedom.findNonFreeness(unshieldedDisjunct, bindingContext)

  lazy val fixedConstantFreedom : Boolean =
    subtree.fixedConstantFreedom && constantFreedom == closingConstantFreedom 

  lazy val stepMeaningful : Boolean = subtree.stepMeaningful && !disjunct.isTrue

  def update(newSubtree : ProofTree,
             newConstantFreedom : ConstantFreedom) : ProofTree =
    if (subtree == newSubtree && constantFreedom == newConstantFreedom)
      this
    else
      WeakenTree(newSubtree, disjunct,
                 vocabulary updateConstantFreedom newConstantFreedom)
  
  def newConstantFreedomForSubtree(cf : ConstantFreedom) : ConstantFreedom = cf

  //////////////////////////////////////////////////////////////////////////////

  override def toString : String =
    "^ " + closingConstraint + " (\\/ " + disjunct + ")\n" + subtree

}
