package ap.proof.goal;

import ap.terfor.{Formula, TermOrder}
import ap.terfor.conjunctions.{Conjunction, Quantifier}
import ap.terfor.preds.{Atom, PredConj, Predicate}
import ap.util.{Debug, Seqs, Logic}

object AddFactsTask {
  
  private val AC = Debug.AC_COMPLEX_FORMULAS_TASK

  /**
   * Return <code>true</code> if <code>f</code> is a formula that can be handled
   * by this task
   */
  def isCoveredFormula(f : Conjunction) : Boolean =
    !f.isFalse &&
    (f.isTrue || f.isLiteral ||
     (f.isNegatedConjunction && f.negatedConjs(0).negatedConjs.isEmpty)) && 
    f.quans.isEmpty

}

class AddFactsTask(_formula : Conjunction, _age : Int)
      extends FormulaTask(_formula, _age) {

  //////////////////////////////////////////////////////////////////////////////
  Debug.assertCtor(AddFactsTask.AC, !formula.isFalse)
  //////////////////////////////////////////////////////////////////////////////

  val priority : Int = age +
                       (if (formula.predicates.isEmpty)
                          -20000
                        else
                          -500)
  /**
   * Perform the actual task (whatever needs to be done with <code>formula</code>)
   */
  protected def doTask(goal : Goal, ptf : ProofTreeFactory) : ProofTree =
    if (formula.isTrue) {
      // then the goal can be closed immediately
      ptf.updateGoal(Conjunction.FALSE, goal)
    } else if (formula.isLiteral) {
      addFacts(Conjunction.negate(formula, goal.order), goal, ptf)
    } else {
      //////////////////////////////////////////////////////////////////////////
      Debug.assertInt(AddFactsTask.AC, formula.isNegatedConjunction)
      //////////////////////////////////////////////////////////////////////////

      val disj = formula.negatedConjs(0)
      //////////////////////////////////////////////////////////////////////////
      // possibly existing quantifiers should have been pulled out and
      // should not occur at this point
      Debug.assertInt(AddFactsTask.AC,
                      disj.quans.isEmpty && disj.negatedConjs.isEmpty)
      //////////////////////////////////////////////////////////////////////////
            
      // the literals of the negated conjunction can be added as facts to
      // the goal
      addFacts(disj, goal, ptf)
    }

  private def addFacts(newFacts : Conjunction,
                       goal : Goal,
                       ptf : ProofTreeFactory) : ProofTree = {
    val updatedFacts = Conjunction.conj(Array(goal.facts, newFacts), goal.order)
    val newTasks =
      unifyAtoms(goal.facts.predConj, newFacts.predConj, goal) ++
      matchClauses(newFacts.predConj, goal)
    ptf.updateGoal(updatedFacts, newTasks, goal)
  }

  private def matchClauses(newFacts : PredConj, goal : Goal) : Seq[PrioritisedTask] =
    NegLitClauseTask.matchClauses(goal.facts.predConj, newFacts,
                                  goal.clauses filter (NegLitClauseTask.isCoveredFormula _),
                                  goal)
  
  private def unifyAtoms(oldFacts : PredConj,
                         newFacts : PredConj,
                         goal : Goal) : Seq[PrioritisedTask] =
    unifyAtoms(oldFacts, true, newFacts, false, goal) ++
    unifyAtoms(oldFacts, false, newFacts, true, goal) ++
    unifyAtoms(newFacts, false, newFacts, true, goal)
  
  private def unifyAtoms(atoms1 : PredConj,
                         // consider the positive or negative literals in atoms1?
                         atoms1pos : Boolean,
                         atoms2 : PredConj,
                         // consider the positive or negative literals in atoms2?
                         atoms2pos : Boolean,
                         goal : Goal) : Seq[PrioritisedTask] = {
    val atoms1seq = if (atoms1pos) atoms1.positiveLits else atoms1.negativeLits
    val atoms2seq = if (atoms2pos) atoms2.positiveLits else atoms2.negativeLits
    if (atoms1seq.size > atoms2seq.size)
      unifyAtoms(atoms2seq, atoms1, atoms1pos, goal)
    else
      unifyAtoms(atoms1seq, atoms2, atoms2pos, goal)
  }
    
  private def unifyAtoms(atoms1 : Seq[Atom],
                         atoms2 : PredConj,
                         // consider the positive or negative literals in atoms2?
                         atoms2pos : Boolean,
                         goal : Goal) : Seq[PrioritisedTask] = {
    var currentPred : Predicate = null
    var currentOtherAtoms : Seq[Atom] = null
    for (atom <- atoms1;
         task <- {
           if (atom.pred != currentPred) {
             currentPred = atom.pred
             currentOtherAtoms = if (atoms2pos)
                                   atoms2 positiveLitsWithPred currentPred
                                 else
                                   atoms2 negativeLitsWithPred currentPred
           }
           unifyAtoms(atom, currentOtherAtoms, goal)
         }) yield task    
  }
  
  private def unifyAtoms(atom : Atom, otherAtoms : Seq[Atom],
                         goal : Goal) : Seq[PrioritisedTask] = {
    for (a2 <- otherAtoms;
         t <- {
                val eqs = atom.unify(a2, goal.order)
                val conj = Conjunction.conj(eqs, goal.order)
                goal.formulaTasks(goal reduceWithFacts conj)
              })
    yield t
  }
  
  /**
   * Create a new <code>FormulaTask</code> by updating the value of
   * <code>formula</code>
   */
  protected def updateFormula(f : Conjunction, goal : Goal) : FormulaTask =
    new AddFactsTask(f, age)

  /**
   * Return <code>true</code> if <code>f</code> is a formula that can be handled
   * by this task
   */
  protected def isCoveredFormula(f : Conjunction) : Boolean =
    AddFactsTask.isCoveredFormula(f)

  protected val name : String = "AddFacts"

}
