package ap.proof.goal;

import ap.basetypes.IdealInt
import ap.terfor.{Term, TermOrder, ConstantTerm}
import ap.terfor.conjunctions.Conjunction
import ap.terfor.inequalities.InEqConj
import ap.terfor.equations.NegEquationConj
import ap.terfor.linearcombination.LinearCombination
import ap.util.{Logic, Debug, Seqs, PlainRange}

import scala.collection.mutable.ArrayBuffer


/**
 * A task for splitting equations in the succedent into two inequalities. This
 * can be necessary if the equation contains the same symbols as other
 * inequalities. Additionally, quantifier-free clauses are split if this is
 * necessary to isolate eliminated constants
 */
case object SplitEqTask extends EagerTask {

  private val AC = Debug.AC_FACTS_TASK

  def apply(goal : Goal, ptf : ProofTreeFactory) : ProofTree = {
    val ac = goal.facts.arithConj
    val order = goal.order
    
    selectConstantToEliminate(goal) match {
      case None => // nothing to be done
                   ptf updateGoal goal
      case Some(constant) => {
        val eqs = ac.negativeEqs filter ((eq) => eq.leadingTerm == constant)
        val clauses = goal.clauses filter ((c) => c.constants contains constant)
        
        // We check whether there are two inequalities of the form
        // <code> constant + t >= 0 & constant + t <= l</code> that could be
        // split, such that <code>l</code> is smaller than the worst-case
        // splitting degree that we can expect from the negated equations
        findSplittableInEqs(constant, goal) match {
          case Some((inEq, d)) if (d < (IdealInt(2) pow (eqs.size + clauses.size))) =>
            splitInEqs(inEq, d, goal, ptf)
          case _ =>
            if (clauses.isEmpty)
              splitEq(eqs.last, goal, ptf)
            else
              splitClause(clauses.first, constant, goal, ptf)
        }
      }
    }
  }

  /**
   * Determine whether this task might causes proof splitting
   */
  def splittingNecessary(goal : Goal) : Boolean =
    selectConstantToEliminate(goal) != None

  /**
   * Split a clause in order to isolate eliminated constants that occur in
   * inequalities
   */
  private def splitClause(c : Conjunction,
                          isolatedConstant : ConstantTerm,
                          goal : Goal, ptf : ProofTreeFactory) : ProofTree = {
    var selectedConjunct : Conjunction = null 
    val otherConjuncts = new ArrayBuffer[Conjunction]
    for (conjunct <- c.elements) {
      if (selectedConjunct == null &&
          (conjunct.constants contains isolatedConstant))
        selectedConjunct = conjunct
      else
        otherConjuncts += conjunct
    }
    
    ////////////////////////////////////////////////////////////////////////////
    Debug.assertInt(SplitEqTask.AC, selectedConjunct != null)
    ////////////////////////////////////////////////////////////////////////////

    BetaFormulaTask.doSplit(selectedConjunct,
                            Conjunction.conj(otherConjuncts, goal.order),
                            goal, ptf)
  }
  
  /**
   * Split an equation in the succedent into two inequalities
   */
  private def splitEq(eq : LinearCombination,
                      goal : Goal,
                      ptf : ProofTreeFactory) : ProofTree = {
    val order = goal.order

    // We have to split the goal directly,
    // because when just adding a formula eq >= 0 & eq <= 0 this will
    // directly be simplified to an equation

    // TODO: directly remove the negated equation
    
    val lowerBound = Conjunction.conj(InEqConj(eq, order), order)
    val upperBound = Conjunction.conj(InEqConj(-eq, order), order)
    val goal1 = ptf.updateGoal(goal.formulaTasks(lowerBound), goal)
    val goal2 = ptf.updateGoal(goal.formulaTasks(upperBound), goal)
    
    ptf.and(Array(goal1, goal2), goal.vocabulary)    
  }

  /**
   * Do a complete case distinction, given that there are pairs of inequalities
   * of the form <code> c + t >= 0 & c + t <= l</code>
   */
  private def splitInEqs(inEq : LinearCombination,
                         distance : IdealInt,
                         goal : Goal,
                         ptf : ProofTreeFactory) : ProofTree = {
    val order = goal.order
    
    val goals = for (d <- PlainRange(distance.intValueSafe + 1)) yield {
      val lc = inEq + IdealInt(-d)
      val newFacts = Conjunction.conj(NegEquationConj(lc, order), order)
      ptf.updateGoal(goal.formulaTasks(newFacts), goal)
    }
    
    ptf.and(goals, goal.vocabulary)
  }
    
  private def findSplittableInEqs(c : ConstantTerm,
                                  goal : Goal) : Option[(LinearCombination, IdealInt)] =
    Seqs.reduceLeft(findAllSplittableInEqs(c, goal),
        (x:(LinearCombination, IdealInt), y:(LinearCombination, IdealInt)) =>
          if ((x _2) < (y _2)) x else y)

  /**
   * Determine whether there are pairs of inequalities of the form
   * <code> c + t >= 0 & c + t <= l</code> that could be split; return all such
   * inequalities, together with the distance between the two bounds
   */
  private def findAllSplittableInEqs(c : ConstantTerm,
                                     goal : Goal) : Seq[(LinearCombination, IdealInt)] = {
    val ac = goal.facts.arithConj
    val inEqs = ac.inEqs
    val order = goal.order

    val res = new ArrayBuffer[(LinearCombination, IdealInt)]
    
    var i = order.findFirstIndex(c, inEqs)
    while (i < inEqs.size && inEqs(i).leadingTerm == c) {
      val lc = inEqs(i)
      
      if (lc.leadingCoeff.isOne) {
        inEqs.findLowerBound(-lc) match {
          case Some(d) => {
            ////////////////////////////////////////////////////////////////////
            Debug.assertInt(SplitEqTask.AC, d.signum < 0)
            ////////////////////////////////////////////////////////////////////
            res += (lc, -d)
          }
          case None => // Nothing
        }
      }
      
      i = i + 1
    }
    
    res
  }

  private def selectConstantToEliminate(goal : Goal) : Option[ConstantTerm] = {
    val facts = goal.facts
    val ac = facts.arithConj
    val order = goal.order

    val res =
      for (lc <- ac.inEqs.firstOption;
           val lt = lc.leadingTerm.asInstanceOf[ConstantTerm];
           if ((goal eliminates lt) && !(facts.predConj.constants contains lt)))
      yield lt
    
    ////////////////////////////////////////////////////////////////////////////
    Debug.assertPost(SplitEqTask.AC,
                     res match {
                       case None =>
                         Seqs.disjoint(goal.eliminatedConstants,
                                       ac.inEqs.constants) ||
                         !Seqs.disjoint(goal.eliminatedConstants,
                                        ac.inEqs.constants,
                                        facts.predConj.constants)
                       case Some(lt) =>
                         (goal eliminates lt) &&
                         !(goal.tasks.taskInfos.constants contains lt) &&
                         ((ac.negativeEqs.constants contains lt) ||
                          Logic.exists(for (c <- goal.clauses.elements)
                                       yield ((c.constants contains lt) &&
                                              c.predicates.isEmpty)))
                     })
    ////////////////////////////////////////////////////////////////////////////
    
    res
  }
    
  private def selectConstantToEliminateXX(goal : Goal) : Option[ConstantTerm] = {
    val ac = goal.facts.arithConj
    val order = goal.order
    
    for (eq <- ac.negativeEqs) {
      if (!eq.isEmpty && (goal eliminates eq.leadingTerm)) {
        val lt = eq.leadingTerm
        ////////////////////////////////////////////////////////////////////////
        Debug.assertInt(SplitEqTask.AC,
                        lt.isInstanceOf[ConstantTerm] &&
                        // this task should only be applied very late, when all
                        // other tasks have already been done
                        !(goal.tasks.taskInfos.constants contains lt.asInstanceOf[ConstantTerm]))
        ////////////////////////////////////////////////////////////////////////
        if (containsLT(lt, ac.inEqs.geqZero, order)) {
          return Some(lt.asInstanceOf[ConstantTerm])
        } else {
          //////////////////////////////////////////////////////////////////////
          Debug.assertInt(SplitEqTask.AC,
                          // the only reason why such an inequality might occur
                          // (i.e., has not been eliminated yet)
                          // is that the leading constant occurs in some
                          // predicate atom or clause
                          (goal.facts.predConj.constants contains lt.asInstanceOf[ConstantTerm]) ||
                          (goal.clauses.constants contains lt.asInstanceOf[ConstantTerm]))
          //////////////////////////////////////////////////////////////////////
        }
      }
    }
    
    None
  }
  
  /**
   * Determine whether a sequence of <code>LinearCombination</code>
   * (sorted in descending order) contains an element with a certain leading
   * term
   */
  private def containsLT(lt : Term,
                         seq : RandomAccessSeq[LinearCombination],
                         order : TermOrder) : Boolean = {
    val i = order.findFirstIndex(lt, seq)
    (i < seq.size) && !seq(i).isEmpty && (seq(i).leadingTerm == lt)
  }

}
