package ap.terfor.preds;

import ap.basetypes.IdealInt
import ap.terfor.linearcombination.LinearCombination
import ap.terfor.equations.EquationConj
import ap.util.{Debug, Logic, Seqs}

object Atom {
  
  private val AC = Debug.AC_PREDICATES

  def apply(pred : Predicate,
            args : Iterator[LinearCombination],
            order : TermOrder) : Atom =
    new Atom (pred, Seqs toRandomAccess args, order)

  def apply(pred : Predicate,
            args : Iterable[LinearCombination],
            order : TermOrder) : Atom =
    new Atom (pred, Seqs toRandomAccess args, order)

}

class Atom private (val pred : Predicate,
                    args : RandomAccessSeq[LinearCombination],
                    val order : TermOrder)
      extends Formula
              with SortedWithOrder[Atom]
              with RandomAccessSeq[LinearCombination] {

  //////////////////////////////////////////////////////////////////////////////
  Debug.assertCtor(Atom.AC,
                   args.size == pred.arity &&
                   Logic.forall(for (lc <- this.elements) yield (lc isSortedBy order)))
  //////////////////////////////////////////////////////////////////////////////
  
  /**
   * Derive equations that describe under which condition this atom describes
   * the same location as that atom (basically, all the argument term have to
   * have the same values)
   */
  def unify(that : Atom, order : TermOrder) : EquationConj =
    if (this.pred == that.pred) {
      if (this.size == 0)
        // no arguments to compare
        EquationConj.TRUE
      else
        EquationConj(for ((arg1, arg2) <- (this.elements zip that.elements))
                     yield LinearCombination.sum(Array((IdealInt.ONE, arg1),
                                                       (IdealInt.MINUS_ONE, arg2)),
                                                 order),
                     order)
    } else {
      EquationConj.FALSE
    }
    
  //////////////////////////////////////////////////////////////////////////////
    
  def length : Int = args.length
      
  def apply(i : Int) : LinearCombination = args(i)
  
  override def elements = args.elements

  //////////////////////////////////////////////////////////////////////////////

  def sortBy(newOrder : TermOrder) : Atom =
    if (isSortedBy(newOrder))
      this
    else
      new Atom (pred,
                (for (a <- args) yield (a sortBy newOrder)).toArray,
                newOrder)
  
  //////////////////////////////////////////////////////////////////////////////
  
  lazy val variables : Set[VariableTerm] =
    Set.empty ++ (for (lc <- this.elements; v <- lc.variables.elements) yield v)

  lazy val constants : Set[ConstantTerm] =
    Set.empty ++ (for (lc <- this.elements; c <- lc.constants.elements) yield c)

  val predicates : Set[Predicate] = Set(pred)
    
  /** Return <code>true</code> if this formula is obviously always true */
  def isTrue : Boolean = false
      
  /** Return <code>true</code> if this formula is obviously always false */
  def isFalse : Boolean = false

  //////////////////////////////////////////////////////////////////////////////

  override def equals(that : Any) : Boolean = that match {
    case that : Atom => (this.pred == that.pred) && (this sameElements that)
    case _ => false
  }
    
  private lazy val hashCodeVal =
    pred.hashCode + Seqs.computeHashCode(this, 911887, 9)

  override def hashCode = hashCodeVal

  override def toString = {
    val strings = for (lc <- this.elements) yield lc.toString
    pred.name +
    (if (strings.hasNext)
       "(" + strings.reduceLeft((s1 : String, s2 : String) => s1 + ", " + s2) + ")"
     else
       "")
  }

}
